########################################
# adapted from: https://github.com/HazyResearch/fm_data_tasks/blob/updates/fm_data_tasks/utils/data_utils.py 
########################################

"""Data utils."""
import os
import copy
import logging
from functools import partial
from pathlib import Path
from typing import Dict, List

import pandas as pd

import utils.contants as constants
from promptsTemplate import *

logger = logging.getLogger(__name__)


ERROR_DETECTION_SPELLING_INSTR =  f"""I have a table with adult data, now I will provide rows of records with the column name: {{column_name}}, please write an appropriate python program that checks if there are any spelling errors. Return "yes" if there is spelling and "no" if there is none."""


def sample_train_data(train: pd.DataFrame, n_rows: int) -> pd.DataFrame:
    """
    Sample train data.

    Used when random sampling points for prompt.
    """
    res = train.sample(n_rows)
    return res

def sample_train_data_stratified(train: pd.DataFrame, n_rows: int) -> pd.DataFrame:
    """
    Sample train data while attempting to maintain the distribution of 'label_str'.
    
    Parameters:
    - train: The training dataset as a pandas DataFrame.
    - n_rows: The total number of rows to sample.
    
    Returns:
    - A pandas DataFrame containing the sampled data.
    """
    # Calculate the number of samples per label based on their distribution in the dataset
    n_labels = train['label_str'].nunique()
    samples_per_label = max(n_rows // n_labels, 1)  # Ensure at least 1 sample per label
    print("Using stratefied sampling! number of samplers per label: ", samples_per_label)
    
    # Adjust n_rows to fit the actual number of samples we can get
    adjusted_n_rows = samples_per_label * n_labels
    
    # Sample from each label
    def sample_from_group(group):
        return group.sample(min(len(group), samples_per_label), replace=True)

    # Group by 'label_str' and apply sampling
    res = train.groupby('label_str', group_keys=False).apply(sample_from_group)
    
    # If adjusted_n_rows < n_rows, sample additional rows randomly to fill up to n_rows
    if len(res) < n_rows:
        additional_samples = n_rows - len(res)
        additional_rows = train.sample(additional_samples)
        res = pd.concat([res, additional_rows], ignore_index=False)
    
    return res

def serialize_row(
    row: pd.core.series.Series,
    column_map: Dict[str, str],
    sep_tok: str,
    nan_tok: str,
) -> str:
    """Turn structured row into string."""
    res = []
    for c_og, c_map in column_map.items():
        if str(row[c_og]) == "nan":
            row[c_og] = nan_tok
        else:
            row[c_og] = f"{row[c_og]}".strip()
        res.append(f"{c_map}: {row[c_og]}".lstrip())
    if len(sep_tok) > 0 and sep_tok != ".":
        sep_tok = f" {sep_tok}"
    return f"{sep_tok} ".join(res)


def serialize_match_pair(
    row: pd.core.series.Series,
    column_mapA: Dict[str, str],
    column_mapB: Dict[str, str],
    add_instruction: bool,
    instruction: str,
    suffix: str,
    prod_name: str,
    sep_tok: str,
    nan_tok: str,
) -> str:
    """Turn structured pair of entities into string for matching."""
    res = (
        f"{prod_name} A is {serialize_row(row, column_mapA, sep_tok, nan_tok)}."
        f" {prod_name} B is {serialize_row(row, column_mapB, sep_tok, nan_tok)}."
        f"{suffix} "
    )
    if add_instruction:
        res = f"{instruction} {res}"
    return res


def serialize_imputation(
    row: pd.core.series.Series,
    column_map: Dict[str, str],
    impute_col: str,
    add_instruction: bool,
    instruction: str,
    suffix: str,
    sep_tok: str,
    nan_tok: str,
) -> str:
    """Turn single entity into string for imputation."""
    assert impute_col not in column_map, f"{impute_col} cannot be in column map"
    # Rename to avoid passing white spaced sep token to serialize_row
    sep_tok_ws = sep_tok
    if len(sep_tok) > 0 and sep_tok != ".":
        sep_tok_ws = f" {sep_tok}"
    res = f"{serialize_row(row, column_map, sep_tok, nan_tok)}{sep_tok_ws}{suffix} "
    if add_instruction:
        res = f"{instruction} {res}"
    return res


def serialize_error_detection_spelling(
    row: pd.core.series.Series,
    add_instruction: bool,
    instruction: str,
    suffix: str,
    sep_tok: str,
    nan_tok: str,
) -> str:
    """Turn single cell into string for error detection."""
    column_map = {row["col_name"]: row["col_name"]}
    res = f"Is there a x spelling error in {serialize_row(row, column_map, sep_tok, nan_tok)}{suffix} "
    if add_instruction:
        res = f"{instruction} {res}"
    return res


def serialize_error_detection(
    row: pd.core.series.Series,
    add_prefix: bool,
    instruction: str,
    suffix: str,
    sep_tok: str,
    nan_tok: str,
) -> str:
    """Turn single cell into string for error detection."""
    column_map = {
        c: c
        for c in row.index
        if str(c) not in ["Unnamed: 0", "text", "col_name", "label_str", "is_clean"]
    }
    entire_row = serialize_row(row, column_map, sep_tok, nan_tok)
    column_map = {row["col_name"]: row["col_name"]}
    res = f"{entire_row}\n\nIs there an error in {serialize_row(row, column_map, sep_tok, nan_tok)}{suffix} "
    if add_prefix:
        res = f"{instruction} {res}"
    return res


def serialize_schema_match(
    row: pd.core.series.Series,
    add_prefix: bool,
    instruction: str,
    suffix: str,
    sep_tok: str,
    nan_tok: str,
) -> str:
    """Turn single cell into string for schema matching."""
    res = f"A is {row['left']}. B is {row['right']}. {suffix} "
    if add_prefix:
        res = f"{instruction}\n\n{res}"
    return res

def serialize_row_simple(row, sep_tok="^", nan_tok='nan'):
    """
    Serialize a DataFrame row into a string.

    Parameters:
    - row: A Pandas Series representing a row in a DataFrame.
    - separator: A string separator used between each value in the row. Defaults to ', '.
    - nan_replacement: A string used to replace NaN values. Defaults to 'N/A'.

    Returns:
    - A string representing the serialized row.
    """
    # Convert each value in the row to string, replace NaN values with the specified replacement
    return sep_tok.join([str(v) if pd.notnull(v) else nan_tok for v in row])


def serialize_row_for_merge(row, prefix, sep_tok, nan_tok):
    # Filter and serialize columns for each part of the merged DataFrame
    filtered_row = row.filter(regex=f'_{prefix}$').rename(lambda x: x[:-2])
    return serialize_row_simple(filtered_row, sep_tok, nan_tok)


def read_blocked_pairs_simple(split_path: str, tableA: pd.DataFrame, tableB: pd.DataFrame, sep_tok='; ', nan_tok='N/A') -> pd.DataFrame:
    labels = pd.read_csv(split_path)
    # Merge the labels with tableA and tableB
    mergedA = pd.merge(labels, tableA, right_on="id", left_on="ltable_id")
    merged = pd.merge(
        mergedA,
        tableB,
        right_on="id",
        left_on="rtable_id",
        suffixes=("_A", "_B"),
    )

    # Format the "text" column by combining serialized parts from tableA and tableB
    merged["text"] = merged.apply(
        lambda row: f"A: {serialize_row_for_merge(row, 'A', sep_tok, nan_tok)}{sep_tok}B: {serialize_row_for_merge(row, 'B', sep_tok, nan_tok)}",
        axis=1,
    )
    # Create the "label_str" column based on the "label" value
    merged["label_str"] = merged["label"].apply(lambda x: "Yes" if x == 1 else "No")

    # Return only the necessary columns
    return merged[["text", "label_str"]]



def read_blocked_pairs(
    split_path: str,
    tableA: pd.DataFrame,
    tableB: pd.DataFrame,
    cols_to_drop: List[str],
    col_renaming: Dict[str, str],
    add_instruction: bool,
    instruction: str,
    suffix: str,
    prod_name: str,
    sep_tok: str,
    nan_tok: str,
) -> pd.DataFrame:
    """Read in pre-blocked pairs with T/F match labels."""
    for c in cols_to_drop:
        tableA = tableA.drop(c, axis=1, inplace=False)
        tableB = tableB.drop(c, axis=1, inplace=False)
    if len(col_renaming) > 0:
        tableA = tableA.rename(columns=col_renaming, inplace=False)
        tableB = tableB.rename(columns=col_renaming, inplace=False)

    column_mapA = {f"{c}_A": c for c in tableA.columns if c != "id"}
    column_mapB = {f"{c}_B": c for c in tableB.columns if c != "id"}

    labels = pd.read_csv(split_path)

    mergedA = pd.merge(labels, tableA, right_on="id", left_on="ltable_id")
    merged = pd.merge(
        mergedA,
        tableB,
        right_on="id",
        left_on="rtable_id",
        suffixes=("_A", "_B"),
    )

    merged["text"] = merged.apply(
        lambda row: serialize_match_pair(
            row,
            column_mapA,
            column_mapB,
            add_instruction,
            instruction,
            suffix,
            prod_name,
            sep_tok,
            nan_tok,
        ),
        axis=1,
    )
    merged["label_str"] = merged.apply(
        lambda row: "Yes\n" if row["label"] == 1 else "No\n", axis=1
    )
    return merged


def read_imputation_single_simple(
    split_path: str,
    impute_col: str,
    sep_tok: str,
    nan_tok: str,
) -> pd.DataFrame:
    """Read in table and create label impute col."""
    table = pd.read_csv(split_path)
    column_map = {c: c for c in table.columns if c != "id" and c != impute_col}
    table["text"] = table.apply(
        lambda row: serialize_row(row=row, column_map=column_map, sep_tok=sep_tok, nan_tok=nan_tok),
        axis=1,
    )
    table["label_str"] = table[impute_col].apply(lambda x: f"{x}\n")
    return table


def read_imputation_single(
    split_path: str,
    impute_col: str,
    cols_to_drop: List[str],
    col_renaming: Dict[str, str],
    add_instruction: bool,
    instruction: str,
    suffix: str,
    sep_tok: str,
    nan_tok: str,
) -> pd.DataFrame:
    """Read in table and create label impute col."""
    table = pd.read_csv(split_path)
    for c in cols_to_drop:
        table = table.drop(c, axis=1, inplace=False)
    if len(col_renaming) > 0:
        table = table.rename(columns=col_renaming, inplace=False)
    column_map = {c: c for c in table.columns if c != "id" and c != impute_col}

    table["text"] = table.apply(
        lambda row: serialize_imputation(
            row,
            column_map,
            impute_col,
            add_instruction,
            instruction,
            suffix,
            sep_tok,
            nan_tok,
        ),
        axis=1,
    )
    table["label_str"] = table[impute_col].apply(lambda x: f"{x}\n")
    return table


def read_data_transformation_single(
        directory_path: str,
        cols_to_drop: List[str],
        col_renaming: Dict[str, str],
        add_instruction: bool,
        instruction: str,
        suffix: str,
        sep_tok: str = '\t',
        nan_tok: str = 'NaN',
    ) -> List[pd.DataFrame]:
        dataframes = []
        for filename in os.listdir(directory_path):
            if filename.endswith('.txt'):
                file_path = os.path.join(directory_path, filename)
                df = pd.read_csv(file_path, sep=sep_tok, names=['Time', 'Hours'])
                # Apply any processing here, similar to your read_imputation_single function
                for c in cols_to_drop:
                    df.drop(c, axis=1, inplace=True)
                if col_renaming:
                    df.rename(columns=col_renaming, inplace=True)
                # Add any specific processing here, you might need to adjust it based on your actual needs
                dataframes.append(df)
        return dataframes


def read_error_detection_single(
    split_path: str,
    table: pd.DataFrame,
    cols_to_drop: List[str],
    col_renaming: Dict[str, str],
    add_instruction: bool,
    instruction: str,
    suffix: str,
    sep_tok: str,
    nan_tok: str,
    spelling: bool,
) -> pd.DataFrame:
    """Read in table and create label impute col."""
    for c in cols_to_drop:
        table = table.drop(c, axis=1, inplace=False)
    if len(col_renaming) > 0:
        table = table.rename(columns=col_renaming, inplace=False)
    # row_id, col_name, is_clean
    labels = pd.read_csv(split_path)

    if spelling:
        merged = pd.merge(labels, table, left_on="row_id", right_index=True)
        merged["text"] = merged.apply(
            lambda row: serialize_error_detection_spelling(
                row,
                add_instruction,
                instruction,
                suffix,
                sep_tok,
                nan_tok,
            ),
            axis=1,
        )
    else:
        merged = table
        merged["text"] = merged.apply(
            lambda row: serialize_error_detection(
                row,
                add_instruction,
                instruction,
                suffix,
                sep_tok,
                nan_tok,
            ),
            axis=1,
        )

    merged["label_str"] = merged.apply(
        lambda row: "No\n" if row["is_clean"] == 1 else "Yes\n", axis=1
    )
    return merged

def read_error_detection_single_simple(
    split_path: str
) -> List[pd.DataFrame]:
    """Read in tables and return a list of dataframes."""
    dataframes = []
    instructions = []
    # List all files in the folder
    for file_name in os.listdir(split_path):
        # Check if the file is a CSV
        if file_name.endswith('.csv'):
            # Extract column name from the file name
            column_name = file_name.split("_")[2].split(".csv")[0]
            
            # Read the CSV file into a DataFrame
            file_path = os.path.join(split_path, file_name)
            df = pd.read_csv(file_path)
            df.columns = df.columns.str.lower()
            # Add the 'column_name' to the DataFrame
            df['text'] = df[column_name]
            
            # Rename 'is_clean' to 'label_str'
            df['label_str'] = df['is_clean'].map({1: 'no', 0: 'yes'})

            
            df.drop('is_clean', axis=1, inplace=True)
            
            # Add the DataFrame to the list
            dataframes.append(df)
            instruction = copy.deepcopy(ERROR_DETECTION_SPELLING_INSTR)
            instructions.append(instruction.format(column_name=column_name))
    return dataframes, instructions

    

def read_schema_match_single(
    split_path: str,
    table: pd.DataFrame,
    cols_to_drop: List[str],
    col_renaming: Dict[str, str],
    add_instruction: bool,
    instruction: str,
    suffix: str,
    sep_tok: str,
    nan_tok: str,
) -> pd.DataFrame:
    """Read in table and create label impute col."""
    file = pd.read_csv(split_path)
    for c in cols_to_drop:
        file = file.drop(c, axis=1, inplace=False)
    if len(col_renaming) > 0:
        file = file.rename(columns=col_renaming, inplace=False)
    # row_id, col_name, is_clean
    # labels = pd.read_csv(split_path)
    # merged = pd.merge(labels, table, left_on="Unnamed: 0", right_index=True)
    file["text"] = file.apply(
        lambda row: serialize_schema_match(
            row,
            add_instruction,
            instruction,
            suffix,
            sep_tok,
            nan_tok,
        ),
        axis=1,
    )
    file["label_str"] = file.apply(
        lambda row: "No\n" if row["label"] == 0 else "Yes\n", axis=1
    )
    return file


def read_transformation_data(data_path, k=3) -> List[pd.DataFrame]:
    """ 
    We read the data transformation dataset here. 
    We first check if there are instructions in the beginning of the file.
    We then split the input output examples to train and test.
    We save the data in to a list of dataframes.
    """
    data_files_sep = {"train": [], "test": [], "instructions": []}  # Dictionary to hold train and test dataframes
    
    for filename in os.listdir(data_path):
        # if "semantic" in filename:
        file_path = os.path.join(data_path, filename)
        
        try:
            with open(file_path, 'r') as file:
                lines = file.readlines()
        except:
            continue 
        # Determine if the first line is an instruction and remove it if so
        instruction = None
        if lines[0].startswith("//"):
            instruction = lines.pop(0).strip()
        
        # Process the remaining lines into input-output pairs
        input_output_pairs = []
        for line in lines:
            if line.strip():
                line_split = line.strip().split('\t\t')
                line_split = [t for t in line_split if t.strip() and not t == '\t' and not t == '\t\t']
                input_output_pairs.append(line_split)
                if len(line_split) > 2:
                    print(file_path)
        # Create a DataFrame from the pairs
        df = pd.DataFrame(input_output_pairs, columns=["text", "label_str"])
        
        # Split the DataFrame into train and test
        train_df = df.iloc[:k]  # The first k entries are for training
        test_df = df.iloc[k:]   # The rest are for testing
        
        # Append the train and test DataFrames to their respective lists in the dictionary
        data_files_sep["train"].append(train_df)
        data_files_sep["test"].append(test_df)
        data_files_sep["instructions"].append(instruction)
        
    return data_files_sep

def sort_list_pairs(pair1, pair2, pair3):
    # Unpack the pairs into separate lists
    list1a, list1b = pair1
    list2a, list2b = pair2
    list3a, list3b = pair3

    # Determine the sorting order based on the second lists
    sort_order = sorted(range(len(list1b)), key=lambda i: (list1b[i], list2b[i], list3b[i]))

    # Sort each first list based on the determined sort order
    sorted_list1a = [list1a[i] for i in sort_order]
    sorted_list2a = [list2a[i] for i in sort_order]
    sorted_list3a = [list3a[i] for i in sort_order]

    # Sort the second lists to make them identical
    sorted_list1b = sorted(list1b)
    sorted_list2b = sorted(list2b)
    sorted_list3b = sorted(list3b)

    # Return the sorted pairs
    return (sorted_list1a, sorted_list1b), (sorted_list2a, sorted_list2b), (sorted_list3a, sorted_list3b)


def read_raw_data_simple(data_dir: str, sep_tok:str, nan_tok:str, k:int=3):
    data_files_sep = {"test": {}, "train": {}, "validation": {}}
    logger.info(f"Reading data from {data_dir}")
    if data_dir not in constants.DATA2TASK:
        raise ValueError(
            f"{data_dir} not one of {constants.DATA2TASK.keys()}. Make sure to set DATASET_PATH."
        )
    task = constants.DATA2TASK[data_dir]
    data_dir_p = Path(data_dir) 
    if task == "entity_matching":
        train_file = data_dir_p / "train.csv"
        valid_file = data_dir_p / "valid.csv"
        test_file = data_dir_p / "test.csv"
        tableA_file = data_dir_p / "tableA.csv"
        tableB_file = data_dir_p / "tableB.csv"

        tableA = pd.read_csv(tableA_file)
        tableB = pd.read_csv(tableB_file)
        label_col = "label"
        data_files_sep["train"] = read_blocked_pairs_simple(train_file, sep_tok=sep_tok, nan_tok=nan_tok, tableA=tableA, tableB=tableB)
        # Read validation
        if valid_file.exists():
            data_files_sep["validation"] = read_blocked_pairs_simple(valid_file,sep_tok=sep_tok, nan_tok=nan_tok, tableA=tableA, tableB=tableB)
        # Read test
        if test_file.exists():
            data_files_sep["test"] = read_blocked_pairs_simple(test_file,sep_tok=sep_tok, nan_tok=nan_tok, tableA=tableA, tableB=tableB)
    elif task == "data_imputation":
        train_file = data_dir_p / "train.csv"
        valid_file = data_dir_p / "valid.csv"
        test_file = data_dir_p / "test.csv"
        label_col = constants.IMPUTE_COLS[data_dir]
        data_files_sep["train"] = read_imputation_single_simple(split_path=valid_file, impute_col=label_col, sep_tok=sep_tok, nan_tok=nan_tok)
        # Read validation
        if valid_file.exists():
            data_files_sep["validation"] = read_imputation_single_simple(split_path=valid_file, impute_col=label_col, sep_tok=sep_tok, nan_tok=nan_tok)
        # Read test
        if test_file.exists():
            data_files_sep["test"] = read_imputation_single_simple(split_path=test_file, impute_col=label_col, sep_tok=sep_tok, nan_tok=nan_tok)
    
    elif task == "error_detection_spelling":
        train_path = data_dir_p / "train_splits_single"
        valid_path = data_dir_p / "valid_splits_single"
        test_path = data_dir_p / "test_splits_single"
        print(train_path)
        label_col = "label_str"
        train_dfs, instructions_train = read_error_detection_single_simple(train_path)
        valid_dfs, instructions_valid = read_error_detection_single_simple(valid_path)
        test_dfs, instructions_test = read_error_detection_single_simple(test_path)
        
        (train_dfs, instructions_train), (valid_dfs, instructions_valid), (test_dfs, instructions_test) = \
            sort_list_pairs((train_dfs, instructions_train), (valid_dfs, instructions_valid), (test_dfs, instructions_test))
        data_files_sep["train"] = train_dfs
        data_files_sep["valid"] = valid_dfs
        data_files_sep["test"] = test_dfs
        data_files_sep["instructions"] = instructions_train
        if instructions_train == instructions_test == instructions_valid:
            print("correct sorted")

    elif task == "data_transformation":
        data_files_sep = read_transformation_data(data_dir_p, k)
        label_col = "label_str"
    else:
        raise ValueError(f"Task {task} not recognized.")
    

    return data_files_sep, label_col


def read_raw_data(
    data_dir: str,
    add_instruction: bool = False,
    task_instruction_idx: int = 0,
    sep_tok: str = ".",
    nan_tok: str = "nan",
    k: int = 3
):
    """Read in data where each directory is unique for a task."""
    data_files_sep = {"test": {}, "train": {}, "validation": {}}
    logger.info(f"Processing {data_dir}")
    if data_dir not in constants.DATA2TASK:
        raise ValueError(
            f"{data_dir} not one of {constants.DATA2TASK.keys()}. Make sure to set DATASET_PATH."
        )
    task = constants.DATA2TASK[data_dir]
    instruction = constants.DATA2INSTRUCT[data_dir]
    suffix = constants.DATA2SUFFIX[data_dir]
    cols_to_drop = constants.DATA2DROPCOLS[data_dir]
    col_renaming = constants.DATA2COLREMAP[data_dir]
    data_dir_p = Path(data_dir)  

    if task == "entity_matching":
        train_file = data_dir_p / "train.csv"
        valid_file = data_dir_p / "valid.csv"
        test_file = data_dir_p / "test.csv"
        tableA_file = data_dir_p / "tableA.csv"
        tableB_file = data_dir_p / "tableB.csv"

        tableA = pd.read_csv(tableA_file)
        tableB = pd.read_csv(tableB_file)

        label_col = "label"
        read_data_func = partial(
            read_blocked_pairs,
            tableA=tableA,
            tableB=tableB,
            cols_to_drop=cols_to_drop,
            col_renaming=col_renaming,
            add_instruction=add_instruction,
            instruction=instruction,
            suffix=suffix,
            prod_name=constants.MATCH_PROD_NAME[data_dir],
            sep_tok=sep_tok,
            nan_tok=nan_tok,
        )
    elif task == "data_imputation":
        train_file = data_dir_p / "train.csv"
        valid_file = data_dir_p / "valid.csv"
        test_file = data_dir_p / "test.csv"
        label_col = constants.IMPUTE_COLS[data_dir]
        read_data_func = partial(
            read_imputation_single,
            impute_col=label_col,
            cols_to_drop=cols_to_drop,
            col_renaming=col_renaming,
            add_instruction=add_instruction,
            instruction=instruction,
            suffix=suffix,
            sep_tok=sep_tok,
            nan_tok=nan_tok,
        )
    elif task == "error_detection_spelling":
        train_file = data_dir_p / "train.csv"
        valid_file = data_dir_p / "valid.csv"
        test_file = data_dir_p / "test.csv"
        table_file = data_dir_p / "table.csv"

        table = pd.read_csv(table_file)
        label_col = "is_clean"
        read_data_func = partial(
            read_error_detection_single,
            table=table,
            cols_to_drop=cols_to_drop,
            col_renaming=col_renaming,
            add_instruction=add_instruction,
            instruction=instruction,
            suffix=suffix,
            sep_tok=sep_tok,
            nan_tok=nan_tok,
            spelling=True,
        )
    elif task == "error_detection":
        train_file = data_dir_p / "train.csv"
        valid_file = data_dir_p / "valid.csv"
        test_file = data_dir_p / "test.csv"
        table_file = data_dir_p / "table.csv"

        table = pd.read_csv(table_file)
        label_col = "is_clean"
        read_data_func = partial(
            read_error_detection_single,
            table=table,
            cols_to_drop=cols_to_drop,
            col_renaming=col_renaming,
            add_instruction=add_instruction,
            instruction=instruction,
            suffix=suffix,
            sep_tok=sep_tok,
            nan_tok=nan_tok,
            spelling=False,
        )
    elif task == "schema_matching":
        train_file = data_dir_p / "train.csv"
        valid_file = data_dir_p / "valid.csv"
        test_file = data_dir_p / "test.csv"
        table_file = data_dir_p / "table.csv"
        label_col = "label"
        table = pd.read_csv(table_file)
        read_data_func = partial(
            read_schema_match_single,
            table=table,
            cols_to_drop=cols_to_drop,
            col_renaming=col_renaming,
            add_instruction=add_instruction,
            instruction=instruction,
            suffix=suffix,
            sep_tok=sep_tok,
            nan_tok=nan_tok,
        )
    elif task == "data_transformation":
        data_files_sep = read_transformation_data(data_dir_p, k)
        label_col = "label_str"
        return data_files_sep, label_col
    else:
        raise ValueError(f"Task {task} not recognized.")

    data_files_sep["train"] = read_data_func(train_file)
    # Read validation
    if valid_file.exists():
        data_files_sep["validation"] = read_data_func(valid_file)
    # Read test
    if test_file.exists():
        data_files_sep["test"] = read_data_func(test_file)
    return data_files_sep, label_col



def read_data(
    data_dir: str,
    class_balanced: bool = False,
    add_instruction: bool = False,
    task_instruction_idx: int = 0,
    max_train_samples: int = -1,
    max_train_percent: float = -1,
    sep_tok: str = ".",
    nan_tok: str = "nan",
):
    """Read in data where each directory is unique for a task."""
    print(os.environ.get("DATASET_PATH", Path("data/datasets").resolve()))
    
    task = constants.DATA2TASK[data_dir]

    data_files_sep, label_col = read_raw_data_simple(
        data_dir=data_dir,
        sep_tok=sep_tok,
        nan_tok=nan_tok,
    )
    task = constants.DATA2TASK[data_dir]
    # Don't class balance on open ended classificiation tasks
    if class_balanced and task != "data_imputation":
        # Class balance sample the train data
        label_cnts = data_files_sep["train"].groupby(label_col).count()
        sample_per_class = label_cnts.min()["text"]
        logger.info(f"Class balanced: train sample per class: {sample_per_class}")
        data_files_sep["train"] = (
            data_files_sep["train"]
            .groupby(label_col, group_keys=False)
            .apply(lambda x: x.sample(sample_per_class, random_state=42))
        )
    # handle different data formats
    if isinstance(data_files_sep["train"], List):
        shuffled_data_files_sep = []
        for train_df in data_files_sep["train"]:
            shuffled_data_files_sep.append(train_df.sample(frac=1, random_state=42).reset_index(drop=True))
        data_files_sep["train"] = shuffled_data_files_sep
    elif isinstance(data_files_sep["train"], pd.DataFrame):
        # Shuffle train data
        data_files_sep["train"] = (
            data_files_sep["train"].sample(frac=1, random_state=42).reset_index(drop=True)
        )
        if max_train_samples > 0:
            orig_train_len = len(data_files_sep["train"])
            if max_train_samples > 1.0:
                raise ValueError("max_train_samples must be between 0 and 1")
            max_examples = int(max_train_samples * orig_train_len)
            data_files_sep["train"] = data_files_sep["train"].iloc[:max_examples]
            logger.info(
                f"Length of {data_dir} train is "
                f"{data_files_sep['train'].shape[0]} from {orig_train_len}"
            )
    return data_files_sep, task


def deserialize_data(data: pd.DataFrame) -> list:
    sampled_examples = []
    for txt, label in zip(data["text"], data["label_str"]):
        sampled_examples.append({"Input": str(txt).strip(), "Output": label.strip() if label else label})

    return sampled_examples


def sample_data_random(train_data: pd.DataFrame, num_examples: int=3) -> Tuple[list, pd.DataFrame]:
    """Get random examples from train data for demonstration. Return a list of dictionaries"""
    prefix_exs_rows = sample_train_data(train_data, num_examples)
    return deserialize_data(prefix_exs_rows), prefix_exs_rows

def sample_data_stratified(train_data: pd.DataFrame, num_examples: int=3) -> Tuple[list, pd.DataFrame]:
    prefix_exs_rows = sample_train_data_stratified(train_data, num_examples)
    return deserialize_data(prefix_exs_rows), prefix_exs_rows


def read_instruction(data_dir: str) -> str:
    """
    Extracts instruction from a text file. Used for data_imputation and entity_matching

    Parameters:
    - data_dir: The directory to the text file containing the instruction.

    Returns:
    - The instruction extracted from the file.
    """
    file_path = os.path.join(data_dir, "instruction.txt")
    try:
        with open(file_path, 'r') as file:
            instruction = file.read().strip()
            # Remove leading '//' if present
            if instruction.startswith("//"):
                instruction = instruction[2:].strip()
            return instruction
    except FileNotFoundError:
        print("The file was not found.")
        return None
    except Exception as e:
        print(f"An error occurred: {e}")
        return None






import logging
import re

from typing import List
from pathlib import Path
from rich.logging import RichHandler

# adapted from https://github.com/HazyResearch/fm_data_tasks/blob/updates/fm_data_tasks/utils/utils.py 
def setup_logger(log_dir: str):
    """Create log directory and logger."""
    Path(log_dir).mkdir(exist_ok=True, parents=True)
    log_path = str(Path(log_dir) / "log.txt")
    handlers = [logging.FileHandler(log_path), RichHandler(rich_tracebacks=True)]
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s [%(module)s] [%(levelname)s] %(message)s",
        handlers=handlers,
    )

def dicts_to_string(list_of_dicts):
    result = ""
    for item in list_of_dicts:
        result += f"Input: \"{item['Input']}\"\nOutput: \"{item['Output']}\"\n"
    return result

def is_within_tolerance(predicted, actual, tolerance=0.021):
    """
    Check if the predicted value is within the tolerance of the actual value.

    Parameters:
    - predicted: The predicted value.
    - actual: The actual (ground truth) value.
    - tolerance: The tolerance level.

    Returns:
    - True if the difference is within the tolerance, False otherwise.
    """
    for pred, act in zip(predicted, actual):
        if  abs(pred - act) > tolerance:
            return False
    return True


def extract_numerical_values(s):
    """
    Extract all numerical values from a given string, including integers, floating point numbers, and fractions.

    Parameters:
    - s: The string containing numerical values.

    Returns:
    - A list of floats found in the string.
    """
    # This regex pattern matches integers, floating point numbers, and fractions.
    pattern = r"[-+]?([0-9]*\.?[0-9]+|[0-9]+\/[0-9]+)"
    matches = re.findall(pattern, s)
    numbers = []
    for match in matches:
        if '/' in match:
            # If the match is a fraction, split it and perform division to get the float equivalent.
            numerator, denominator = map(float, match.split('/'))
            number = numerator / denominator
        else:
            # Otherwise, directly convert the match to float.
            number = float(match)
        numbers.append(number)
    return numbers if numbers else None


def evaluate_numerical_values(pred, label):
    """
    Check if the prediction contains numerical values. 
    If yes, then check if difference between label and pred is within certain tolerance.
    """
    crc = None
    num_pred_try = None
    num_gt_try = None
    if isinstance(pred, str):
        num_pred_try = extract_numerical_values(pred)
    elif isinstance(pred, int) or isinstance(pred, float):
        num_pred_try = [pred]

    if isinstance(label, str):
        num_gt_try = extract_numerical_values(label)
    elif isinstance(label, int) or isinstance(label, float):
        num_gt_try = [label]
    if num_pred_try and num_gt_try:
        # num_label = extract_numerical_values(label)
        if is_within_tolerance(num_pred_try, num_gt_try):
            crc = True
        else:
            crc = False
    else:
        crc = False
    return crc
    

def compute_metrics(preds: List, golds: List, task: str):
    """Compute metrics."""
    mets = {"tp": 0, "tn": 0, "fp": 0, "fn": 0, "crc": 0, "total": 0}
    for pred, label in zip(preds, golds):
        if label:
            if isinstance(label, str):
                label = label.strip().lower()
        if isinstance(pred, tuple) or isinstance(pred, List):
            pred = ' '.join(str(element) for element in pred)
        elif pred == None:
            pred = ""
        if label == None:
            label = ""
        if isinstance(pred, str):
            pred = pred.strip().lower()
        mets["total"] += 1
        if task in {
            "data_imputation",
            "entity_matching",
            "data_transformation",
            "error_detection_spelling"
        }:
            if pred == label:
                crc = True
            else:
                # if they are not equal, check if it is numerical values within tolerance
                crc = evaluate_numerical_values(pred, label)
        else:
            raise ValueError(f"Unknown task: {task}")
        # Measure equal accuracy for generation
        if crc:
            mets["crc"] += 1
        if label == "yes":
            if crc:
                mets["tp"] += 1
            else:
                mets["fn"] += 1
        elif label == "no":
            if crc:
                mets["tn"] += 1
            else:
                mets["fp"] += 1

    prec = mets["tp"] / max(1, (mets["tp"] + mets["fp"]))
    rec = mets["tp"] / max(1, (mets["tp"] + mets["fn"]))
    if mets["total"] == 0:
        acc = 0.
    else:
        acc = mets["crc"] / mets["total"]
    f1 = 2 * prec * rec / max(1, (prec + rec))
    return prec, rec, acc, f1


import ast
import openai
import sys
import logging
import multiprocessing
sys.path.insert(0, '')

from src.tableswift.utils import function_utils
from src.tableswift.utils.formatter_utils import format_model_output_python
from .base_generator import CodeGenerator
from pydantic import BaseModel, Field

print(function_utils)

logger = logging.getLogger(__name__)

class PythonResponse(BaseModel):
    reasoning: str
    python_code: str

class PythonResponse_DataRouter(BaseModel):
    reasoning: str
    solving_task_code: str
    validate_code: str

class PythonGenerator(CodeGenerator):
    def __init__(self, task: str, llm: str, use_data_router:bool, api_key:str) -> None:
        super().__init__(task, llm, api_key)
        self.lang = "python"
        self.use_data_router = use_data_router

    def get_response(self, messages):
        """
        We use instructor and pre-defined response model
        """
        print("here is the complete messages, ", messages)
        response_model = PythonResponse if not self.use_data_router else PythonResponse_DataRouter
        structured_output = self.call_llm(messages, response_model)
        if self.llm in self.ollama_models or self.llm in self.together_models or self.llm in self.openrouter_models:
            structured_output = format_model_output_python(structured_output)
            print("Structured output after formatting: ", structured_output)
        return structured_output

    def pre_process(self, input: str):
        if isinstance(input, str):
            input = input.replace("'", "''")
        return input
    
    def post_process(self, structured_output):
        """
        Post process the structured output, return a dictionary
        This is the same for both code generators
        """
        result = {"resoning": None, "transformation_code": None, "validation_code": None}
        if self.llm in self.ollama_models or self.llm in self.together_models or self.llm in self.openrouter_models:
            result["transformation_code"] = structured_output
            return result
        if not self.use_data_router:
            try:
                reasoning = structured_output.reasoning
                result["reasoning"] = reasoning
                transformation_code = structured_output.python_code
                if transformation_code:
                    transformation_code = transformation_code.replace('\\\\', '\\')
                    result["transformation_code"] = transformation_code
            except:
                pass
            return result
        else:
            try:
                reasoning = structured_output.reasoning
                result["reasoning"] = reasoning
                transformation_code = structured_output.solving_task_code
                validation_code = structured_output.validate_code
                if transformation_code:
                    transformation_code = transformation_code.replace('\\\\', '\\')
                    result["transformation_code"] = transformation_code
                if validation_code:
                    validation_code = validation_code.replace('\\\\', '\\')
                    result["validation_code"] = validation_code
            except:
                pass
            return result
    
    # def execute_function_string(self, fn, input_string, fn_type="transformation"):
    #     # TODO: parse to get the function name.
    #     # Prepare a custom namespace for executing the dynamic code. This can be an empty dict.
    #     if fn_type == "transformation":
    #         if self.task == "data_transformation":
    #             fn_name = "string_transformation"
    #         elif self.task == "entity_matching":
    #             fn_name = "entity_matcher"
    #         elif self.task == "data_imputation":
    #             fn_name = "impute_data"
    #         elif self.task == "error_detection_spelling":
    #             fn_name = "detect_error"
    #         else:
    #             raise NotImplementedError
    #     elif fn_type == "validation":
    #         fn_name = "validate"
    #     else:
    #         raise NotImplementedError
        
    #     namespace = {}
    #     # Execute the dynamic code. This will define the function within the 'namespace'.
    #     exec(fn, namespace)
    #     # Access the function from the namespace and call it
    #     func = namespace[fn_name]
    #     result = func(input_string) 
    #     return result
    
    # def execute(self, fn, test_data, fn_type="transformation"):
    #     try:
    #         _ = self.execute_function_string(fn, test_data[0]['Input'], fn_type)
    #     except Exception as e:
    #         return False, f"Error: {str(e)}"
    #     return True, "Execution successful."

    

    def execute_function_string(self, fn, input_string, fn_type="transformation"):
        """
        Executes a Python function string directly (no timeout logic here).
        """
        # Determine the function name based on the task and type
        if fn_type == "transformation":
            fn_name = {
                "data_transformation": "string_transformation",
                "entity_matching": "input_output_converter",
                "data_imputation": "input_output_converter",
                "error_detection_spelling": "detect_error",
            }.get(self.task, None)
        elif fn_type == "validation":
            fn_name = "validate"
        if not fn_name:
            raise NotImplementedError(f"Task '{self.task}' not supported.")

        # Prepare a custom namespace for executing the dynamic code
        namespace = {}
        exec(fn, namespace)  # Define the function in the namespace
        func = namespace[fn_name]  # Access the function by name
        return func(input_string)  # Call the function with input data
    
    def target_function(self, queue, fn, input_string, fn_type):
        """
        Executes the function string and puts the result and message in a queue.
        """
        try:
            result = self.execute_function_string(fn, input_string, fn_type=fn_type)
            queue.put((result, "Execution successful."))
        except Exception as e:
            queue.put((None, f"Error: {str(e)}"))

    def execute(self, fn, test_data, fn_type="transformation", timeout=5):
        """
        Wraps `execute_function_string` with timeout logic and returns both result and message.
        """
        if isinstance(test_data, list):
            input_string = test_data[0]['Input']
        elif isinstance(test_data, str):
            input_string = test_data
        queue = multiprocessing.Queue()
        process = multiprocessing.Process(
            target=self.target_function,
            args=(queue, fn, input_string, fn_type)
        )

        process.start()
        process.join(timeout)

        if process.is_alive():
            process.terminate()  # Forcefully terminate the process
            process.join()
            return None, f"Error: Execution timed out after {timeout} seconds."

        # Retrieve the result and message from the queue
        if not queue.empty():
            return queue.get()
        return None, "Error: Execution failed: Unknown error."
    
    def evaluate(self, fn, test_data):
        inputs = [item["Input"] for item in test_data]
        outputs = [item["Output"] for item in test_data]
        predicted_outputs = []
        for input, output in zip(inputs, outputs):
            try:
                result, _ = self.execute(fn, input)
            except Exception as error:
                print(f"Error message: {error}")
                result = "Not excutable"
            predicted_outputs.append(result)
            print("pred: {}, gt: {}".format(result, output))
        # acc = calculate_accuracy(predicted=predicted_outputs, ground_truth=outputs)
        metrics = function_utils.compute_metrics(preds=predicted_outputs, golds=outputs, task=self.task)
        if self.task in [
                "data_imputation",
                "data_transformation",
                
            ]:
            acc = metrics[2]
        elif self.task in ["entity_matching", "error_detection_spelling"]:
            acc = metrics[3]
        print("The accuacry/f1 of the generated function is ", acc)
        return acc, predicted_outputs
    
    def route_data(self, v_fn, test_data):
        """
        Route the data given the response from the validation function
        """
        valid_data = []
        invalid_data = []
        if v_fn:
            inputs = [item["Input"] for item in test_data]
            outputs = [item["Output"] for item in test_data]
            for input, output in zip(inputs, outputs):
                try:
                    decision, _ = self.execute(v_fn, input, "validation")
                    logger.info(f"======> Decision for input {input} is {decision} <======")
                    if decision:
                        valid_data.append({"Input": input, "Output": output})
                    else:
                        invalid_data.append({"Input": input, "Output": output})
                except Exception as error:
                    # if the validation function is not excutable, we treat it as valid data
                    print(f"Error message: {error}")
                    valid_data.append({"Input": input, "Output": output})
        else:
            valid_data = test_data
        return valid_data, invalid_data
    
    def pipeline_backbone(self, messages, examples, supervision_data, depth):
        """
        Pipeline for getting a valid transformation function
        """
        func = None
        num_iter = 1
        functions_stack = []
        result = None
        while num_iter < depth:
            # getting the first function
            response = self.get_response(messages)
            result = self.post_process(response)
            if result["transformation_code"]:
                func = result["transformation_code"]
                decision, msg, acc = self.validate(func, examples, supervision_data)
                # for debugging
                print("The decision is ", decision)
                print("The message is ", msg)
                print("The accuracy is ", acc)

                if "not" in decision:
                    if not decision == "not_excutable":
                        functions_stack.append((func, acc))
                    # if this is the last try and the function runs then we return
                    if (num_iter == depth - 1):
                        if functions_stack:
                            # sort the function stack
                            sorted_func_stack = sorted(functions_stack, key = lambda x: x[1], reverse=True)
                            return (*sorted_func_stack[0], result)
                    # add error message and re-prompt
                    retry_msg = self.prompts_set["TASK_RETRY"][0].copy()
                    retry_msg['content'] = retry_msg['content'].format(error_message=msg)
                    messages.append(retry_msg)     
                else:
                    return (func, acc, result)           
            else:
                decision = "not parsable"
                msg = "The function calling return not parsable response, trying again!"
                acc = 0
                 # add error message and re-prompt
                retry_msg = self.prompts_set["TASK_RETRY"][0].copy()
                retry_msg['content'] = retry_msg['content'].format(error_message=msg)
                messages.append(retry_msg)   
            num_iter += 1
        return ("No function can be generated, please provide different demonstration!", float("-inf"), result)

    def pipeline_data_router(self, messages, examples, supervision_data, depth):
        """
        Pipeline with data router
        """
        # first prioritize the transformation code, after getting a valid transformation code, we then validate if the validation code is valid
        # for now we only valid the validation function but not re-generate
        # TODO: in the future we can re-generate the validation function
        t_func = None
        t_func, acc, result = self.pipeline_backbone(messages, examples, supervision_data, depth)
        if not t_func:
            return t_func, acc, None
        # now we validate the validation function
        v_func = result["validation_code"]
        if not v_func:
            return t_func, acc, None
        runnable, msg = self.execute(v_func, examples, "validation")
        if runnable:
            return t_func, acc, v_func
        else:
            return t_func, acc, None
        
    def pipeline(self, instruction, examples, supervision_data=None, depth: int=5):
        """
        The sequence of action for generating the final function is: generate function -> validate function -> correct function if necessary.
        """
        result = {"transformation_code": None, "acc": None, "validation_code": None}
        # Define the messages
        messages, _ = self.formulate_prompt(instruction=instruction, examples=function_utils.dicts_to_string(examples))   
        if not self.use_data_router:
            code, acc, _ = self.pipeline_backbone(messages, examples, supervision_data, depth)
            result["transformation_code"] = code
            result["acc"] = acc
        else:
            # use pipeline that contains data router
            code, acc, v_code = self.pipeline_data_router(messages, examples, supervision_data, depth)
            result["transformation_code"] = code
            result["acc"] = acc
            result["validation_code"] = v_code
        return result


# DuckDB NL2SQL formatter

def format_model_output_sql(output_sql: str) -> str:
        def clean_code_block(block: str) -> str:
            """Clean a code block by removing markdown syntax and extra whitespace."""
            # Remove markdown indicators and common SQL prefixes
            cleaned = (block
                       .replace('```sql\n', '')
                       .replace('```duckdb\n', '')
                       .replace('```\n', '')
                       .replace('```', '')
                       .strip())

            return cleaned

        def ensure_semicolon(sql: str) -> str:
            """Ensure the SQL query ends with exactly one semicolon."""
            sql = sql.strip()
            # Remove any existing trailing semicolons
            while sql.endswith(';'):
                sql = sql[:-1].strip()
            # Add back exactly one semicolon
            return sql + ";"

        # First, try to find SQL-specific code blocks
        sql_blocks = []
        start_pos = 0
        while True:
            start = output_sql.find('```sql', start_pos)
            if start == -1:
                start = output_sql.find('```duckdb', start_pos)
            if start == -1:
                break

            end = output_sql.find('```', start + 4)
            if end == -1:
                break

            sql_blocks.append(output_sql[start:end+3])
            start_pos = end + 3

        # If SQL blocks found, use the last one
        if sql_blocks:
            return ensure_semicolon(clean_code_block(sql_blocks[-1])).replace('```sql','').replace('```','').strip()

        # If no SQL blocks, look for generic code blocks
        generic_blocks = []
        start_pos = 0
        while True:
            start = output_sql.find('```', start_pos)
            if start == -1:
                break

            end = output_sql.find('```', start + 3)
            if end == -1:
                break

            block = output_sql[start:end+3]
            # Skip if this is actually an SQL block (we already handled those)
            if not block.startswith('```sql') and not block.startswith('```duckdb'):
                generic_blocks.append(block)
            start_pos = end + 3

        # If generic blocks found, use the last one
        if generic_blocks:
            return ensure_semicolon(clean_code_block(generic_blocks[-1])).replace('```sql','').replace('```','').strip()

        # If no code blocks found at all, take everything up to first semicolon
        semicolon_pos = output_sql.find(';')
        if semicolon_pos != -1:
            return ensure_semicolon(output_sql[:semicolon_pos].strip()).replace('```sql','').replace('```','').strip()

        # If no semicolon found, use the entire text
        return ensure_semicolon(output_sql.strip()).replace('```sql','').replace('```','').strip()


def format_model_output_python(output_python: str) -> str:
    def clean_code_block(block: str) -> str:
        """Clean a code block by removing markdown syntax and extra whitespace."""
        # Remove markdown indicators and common Python prefixes
        cleaned = (block
                    .replace('```python\n', '')
                    .replace('```Python\n', '')
                    .replace('```\n', '')
                    .replace('```', '')
                    .strip())

        return cleaned

        # def ensure_semicolon(python_code: str) -> str:
        #     """Ensure the Python ends with exactly one semicolon."""
        #     python_code = python_code.strip()
        #     # Remove any existing trailing semicolons
        #     while python_code.endswith(';'):
        #         python_code = python_code[:-1].strip()
        #     # Add back exactly one semicolon
        #     return python_code + ";"

        # First, try to find Python-specific code blocks
    python_blocks = []
    start_pos = 0
    while True:
        start = output_python.find('```python', start_pos)
        if start == -1:
            start = output_python.find('```Python', start_pos)
        if start == -1:
            break

        end = output_python.find('```', start + 4)
        if end == -1:
            break

        python_blocks.append(output_python[start:end+3])
        start_pos = end + 3

    # If SQL blocks found, use the last one
    if python_blocks:
        return (clean_code_block(python_blocks[-1])).replace('```Python','').replace('```','').strip()

    # If no SQL blocks, look for generic code blocks
    generic_blocks = []
    start_pos = 0
    while True:
        start = output_python.find('```', start_pos)
        if start == -1:
            break

        end = output_python.find('```', start + 3)
        if end == -1:
            break

        block = output_python[start:end+3]
        # Skip if this is actually an SQL block (we already handled those)
        if not block.startswith('```python') and not block.startswith('```Python'):
            generic_blocks.append(block)
        start_pos = end + 3

    # If generic blocks found, use the last one
    if generic_blocks:
        return clean_code_block(generic_blocks[-1]).replace('```python','').replace('```','').strip()

    # # If no code blocks found at all, take everything up to first semicolon
    # semicolon_pos = output_python.find(';')
    # if semicolon_pos != -1:
    #     return output_sql[:semicolon_pos].strip().replace('```sql','').replace('```','').strip()

    # If no semicolon found, use the entire text
    return output_python.strip().replace('```python','').replace('```','').strip()


# model_output = """```python
# def string_transformation(input_string):
#     input_value = int(input_string.replace(' mg', ''))
#     output_value = input_value * 2
#     return f"{output_value} ml"
# ```"""
# print("here is the result")
# print(format_model_output_python(model_output))


import os
import openai
from openai import OpenAI
from together import Together
import instructor
from pydantic import BaseModel
from ollama import chat
from ollama import ChatResponse

from abc import ABC
from ..promptsTemplate import get_prompts_set, PromptTemplate

class CodeGenerator(ABC):
    """
    All functions here are programming language independent.
    """
    def __init__(self, task: str, llm: str, api_key: str) -> None:
        super().__init__()
        self.lang = None
        self.use_data_router = None
        self.task = task
        # it can be ["gpt-4-turbo", "llama3.2", "qwen2.5-coder"]
        self.llm = llm
        self.prompts_set = None
        self.api_key = api_key

        # pre-defined categories
        self.openai_models = {"gpt-4": "gpt-4-turbo", "gpt-4o-mini": "gpt-4o-mini"}
        # openrouter models
        self.openrouter_models = {"deepseek-r1": "deepseek/deepseek-r1:free"}
        # llama3.2: llama3.2-3b-instruct, qwen2.5-coder: qwen2.5-coder-7b
        # ollama is completely local and free
        self.ollama_models = {"llama3.2-3b": "llama3.2", "qwen2.5-coder-7b": "qwen2.5-coder"}
        # llama3.1-405b
        self.together_models = {"llama3.1-405b": "meta-llama/Meta-Llama-3.1-405B-Instruct-Turbo", 
                                "llama3.1-70b": "meta-llama/Meta-Llama-3.1-70B-Instruct-Turbo",
                                "qwen2.5-coder-32b": "Qwen/Qwen2.5-Coder-32B-Instruct",
                                "llama3.1-8b": "meta-llama/Meta-Llama-3.1-8B-Instruct-Turbo",
                                "llama3.2-3b": "meta-llama/Llama-3.2-3B-Instruct-Turbo"}
    
    def formulate_prompt(self, instruction: str, examples: str):
        prompts_set = get_prompts_set(self.lang, use_data_router=self.use_data_router)
        if self.task == "data_transformation":
            template_name = "STRING_TRANSFORMATION"
        elif self.task == "entity_matching":
            template_name = "ENTITY_MATCHING"
        elif self.task == "data_imputation":
            template_name = "DATA_IMPUTATION"
        elif self.task == "error_detection_spelling":
            template_name = "ERROR_DETECTION_SPELLING"
        else:
            raise NotImplementedError
        prefix = prompts_set["TASK_PREFIX"]
        if template_name not in prompts_set:
            raise NotImplementedError
        else:
            template = prompts_set[template_name]
        prompts = prefix + template
        prompt_template = PromptTemplate(prompt=prompts)
        prompts = []
        for prompt_t in prompt_template.prompt:
            prompt = prompt_t.copy()
            if 'user' in prompt_t['role']:
                print("Found user")
                prompt['content'] = prompt['content'].format(instruction=instruction, examples=examples)
            prompts.append(prompt)
        self.prompts_set = prompts_set
        return prompts, prompt_template
    

    def call_llm(self, messages, response_model):
        print(self.llm)
        if self.llm in self.openrouter_models:
            model_name = self.openrouter_models[self.llm]
            client = OpenAI(
                base_url="https://openrouter.ai/api/v1",
                api_key=self.api_key,
            )
            # Extract structured data from natural language
            response = client.chat.completions.create(
                model=model_name,
                messages=messages)
            structured_output = response.choices[0].message.content
            print("structured output   ", structured_output)
        elif self.llm in self.openai_models:
            model_name = self.openai_models[self.llm]
            client = instructor.from_openai(client = openai.OpenAI(api_key=self.api_key),)

            # Extract structured data from natural language
            structured_output = client.chat.completions.create(
                model=model_name,
                response_model=response_model,
                messages=messages,
            )
        elif self.llm in self.ollama_models:
            # not using instructor here because it doesn't work well with smaller models
            # enables `response_model` in create call
            # client = instructor.from_openai(
            #     OpenAI(
            #         base_url="http://localhost:11434/v1",
            #         api_key="ollama",  # required, but unused
            #     ),
            #     mode=instructor.Mode.JSON,
            # )
            # resp = client.chat.completions.create(
            #     model="llama3.2",
            #     messages=messages,
            #     response_model=response_model,
            # )
            model_name = self.ollama_models[self.llm]
            response: ChatResponse = chat(model=model_name, messages=messages)
            structured_output = response['message']['content']
            print(f"structured output   {structured_output}")
        elif self.llm in self.together_models:
            model_name = self.together_models[self.llm]
            client = Together(api_key=self.api_key)

            response = client.chat.completions.create(
                model=model_name,
                messages=messages,
            )
            structured_output = response.choices[0].message.content
            print(f"structured output   {structured_output}")
        else:
            raise NotImplementedError
        
        return structured_output

    def post_process():
        pass

    def execute():
        """
        Taking a piece of code string, try to execute it. Return true if it is executable.
        """
        pass

    def validate(self, fn, test_data, supervision_data, threshold_sup=10., threshold=0.51):
        """
        We validate function by checking:
        1. If the function runs.
        2. If the function get ~100% [self-defined] accuracy on the demonstrations.
        3. [Optional]If the function is generalizable by prompting LLMs. (currently not supported).
        """
        runnable, msg = self.execute(fn, test_data)
        if not runnable:
            error_message = f"The generated code {fn} is not excutable, error message is {msg}. Please fix the code."
            return "not_excutable", error_message, 0
        acc, _ = self.evaluate(fn, test_data)
        # we don't expect accuracy to be 100% on currency exchange tasks
        if acc < threshold:
            error_message = f"The generated code {fn} only achieves {acc} accuracy on provided examples. Please provide different code to achieve higher accuracy."
            return "not_fit", error_message, acc
        if supervision_data and not len(supervision_data) == 0:
            acc_sup, _ = self.evaluate(fn, supervision_data)
            if acc_sup < threshold_sup:
                error_message = f"The generated code {fn} is overfitting on the provided example, please reconsider the intention of this task."
                return "not_generalizable", error_message, acc
        return "validated", "", acc

    def evaluate():
        pass

    def pipeline():
        # here we have the logic of generate, validate, rank and omit
        pass